package ru.dbotthepony.kstarbound.lua.bindings

import org.classdump.luna.ByteString
import org.classdump.luna.Table
import org.classdump.luna.runtime.ExecutionContext
import ru.dbotthepony.kommons.collect.map
import ru.dbotthepony.kommons.collect.toList
import ru.dbotthepony.kstarbound.math.vector.Vector2d
import ru.dbotthepony.kstarbound.lua.LuaEnvironment
import ru.dbotthepony.kstarbound.lua.from
import ru.dbotthepony.kstarbound.lua.get
import ru.dbotthepony.kstarbound.lua.iterator
import ru.dbotthepony.kstarbound.lua.luaFunction
import ru.dbotthepony.kstarbound.lua.set
import ru.dbotthepony.kstarbound.lua.toPoly
import ru.dbotthepony.kstarbound.lua.toVector2d
import ru.dbotthepony.kstarbound.util.valueOf
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.physics.CollisionType
import java.util.EnumSet
import java.util.function.Predicate

fun provideWorldBindings(self: World<*, *>, lua: LuaEnvironment) {
	val callbacks = lua.newTable()
	lua.globals["world"] = callbacks

	callbacks["flyingType"] = luaFunction {
		returnBuffer.setTo(self.sky.flyingType.jsonName)
	}

	callbacks["magnitude"] = luaFunction { arg1: Table, arg2: Table? ->
		if (arg2 == null) {
			returnBuffer.setTo(toVector2d(arg1).length)
		} else {
			returnBuffer.setTo(self.geometry.diff(toVector2d(arg1), toVector2d(arg2)).length)
		}
	}

	callbacks["polyContains"] = luaFunction { poly: Table, position: Table ->
		returnBuffer.setTo(self.geometry.polyContains(toPoly(poly), toVector2d(position)))
	}

	callbacks["xwrap"] = luaFunction { position: Any ->
		if (position is Number) {
			returnBuffer.setTo(self.geometry.x.cell(position.toDouble()))
		} else {
			returnBuffer.setTo(from(self.geometry.wrap(toVector2d(position))))
		}
	}

	callbacks["ywrap"] = luaFunction { position: Any ->
		if (position is Number) {
			returnBuffer.setTo(self.geometry.y.cell(position.toDouble()))
		} else {
			returnBuffer.setTo(from(self.geometry.wrap(toVector2d(position))))
		}
	}

	callbacks["nearestTo"] = luaFunction { pos0: Any, pos1: Any ->
		if (pos1 is Number) {
			val source = if (pos0 is Number) pos0.toDouble() else toVector2d(pos0).x
			returnBuffer.setTo(self.geometry.x.nearestTo(source, pos1.toDouble()))
		} else {
			val source = if (pos0 is Number) Vector2d(pos0.toDouble()) else toVector2d(pos0)
			returnBuffer.setTo(self.geometry.nearestTo(source, toVector2d(pos1)))
		}
	}

	callbacks["rectCollision"] = luaFunction { rect: Table, collisions: Table? ->
		if (collisions == null) {
			returnBuffer.setTo(self.collide(toPoly(rect), Predicate { it.type.isSolidCollision }).findAny().isPresent)
		} else {
			val actualCollisions = EnumSet.copyOf(collisions.iterator().map { CollisionType.entries.valueOf((it.value as ByteString).decode()) }.toList())
			returnBuffer.setTo(self.collide(toPoly(rect), Predicate { it.type in actualCollisions }).findAny().isPresent)
		}
	}
}
